@@ -1,24 +1,24 @@
 #! /usr/bin/env python3
 
-import os
-import io
-import errno
 import argparse
-import logging
-import threading
+import configparser
+import errno
 import filecmp
+import io
+import json
+import logging
+import os
 import shutil
 import subprocess
-import json
-import configparser
 import sys
+import threading
 
 from collections import OrderedDict
 
 try:
     import git
     import semantic_version as semver # lgtm[py/unused-import]
-except ImportError:
+except ImportError as error:
     print("error: zkg failed to import one or more dependencies:\n"
           "\n"
           "* GitPython:        https://pypi.org/project/GitPython\n"
@@ -26,15 +26,20 @@ except ImportError:
           "\n"
           "If you use 'pip', they can be installed like:\n"
           "\n"
-          "     pip3 install GitPython semantic-version",
+          "     pip3 install GitPython semantic-version\n"
+          "\n"
+          "Also check the following exception output for possible alternate explanations:\n\n"
+          "{}: {}".format(type(error).__name__, error),
           file=sys.stderr)
     sys.exit(1)
 
-# For Zeek-bundled installation, explictly add the Python path we've
-# installed ourselves under, so we find the zeekpkg module:
+# For Zeek-bundled installation, prepend the Python path of the Zeek
+# installation. This ensures we find the matching zeekpkg module
+# first, avoiding potential conflicts with installations elsewhere on
+# the system.
 ZEEK_PYTHON_DIR = '@PY_MOD_INSTALL_DIR@'
 if os.path.isdir(ZEEK_PYTHON_DIR):
-    sys.path.append(os.path.abspath(ZEEK_PYTHON_DIR))
+    sys.path.insert(0, os.path.abspath(ZEEK_PYTHON_DIR))
 else:
     ZEEK_PYTHON_DIR = None
 
@@ -65,7 +70,11 @@ if not os.path.isdir(ZEEK_ZKG_STATE_DIR):
 # The default package source we fall back to as needed
 ZKG_DEFAULT_SOURCE = 'https://github.com/zeek/packages'
 
+# The default package template
+ZKG_DEFAULT_TEMPLATE = 'https://github.com/zeek/package-template'
+
 from zeekpkg._util import (
+    delete_path,
     make_dir,
     make_symlink,
     find_program,
@@ -75,6 +84,13 @@ from zeekpkg._util import (
 from zeekpkg.package import (
     TRACKING_METHOD_VERSION,
 )
+from zeekpkg.template import (
+    LoadError,
+    Template,
+)
+from zeekpkg.uservar import (
+    UserVar,
+)
 
 import zeekpkg
 
@@ -103,7 +119,7 @@ def confirmation_prompt(prompt, default_to_yes=True):
     return False
 
 
-def prompt_for_user_vars(manager, config, configfile, force, pkg_infos):
+def prompt_for_user_vars(manager, config, configfile, args, pkg_infos):
     answers = {}
 
     for info in pkg_infos:
@@ -114,45 +130,15 @@ def prompt_for_user_vars(manager, config, configfile, force, pkg_infos):
             print_error(str.format('error: malformed user_vars in "{}"', name))
             sys.exit(1)
 
-        if not requested_user_vars:
-            continue
+        for uvar in requested_user_vars:
+            try:
+                answers[uvar.name()] = uvar.resolve(name, config, args.user_var, args.force)
+            except ValueError:
+                print_error(str.format('error: could not determine value of user variable "{}",'
+                                       ' provide via environment or --user-var', uvar.name()))
+                sys.exit(1)
 
-        for key, value, desc in requested_user_vars:
-            from_env = False
-            default_value = os.environ.get(key)
-
-            if default_value:
-                from_env = True
-            else:
-                if config.has_section('user_vars'):
-                    v = config.get('user_vars', key, fallback=None)
-
-                    if v:
-                        default_value = v
-                    else:
-                        default_value = value
-                else:
-                    default_value = value
-
-            if force:
-                answers[key] = default_value
-            else:
-                if from_env:
-                    print(str.format(
-                        '{} will use value of {} ({}) from environment: {}',
-                        name, key, desc, default_value))
-                    answers[key] = default_value
-                else:
-                    prompt = '{} asks for {} ({}) ? [{}] '.format(
-                        name, key, desc, default_value)
-                    response = input(prompt)
-
-                    if response:
-                        answers[key] = response
-                    else:
-                        answers[key] = default_value
-
-    if not force and answers:
+    if not args.force and answers:
         for key, value in answers.items():
             if not config.has_section('user_vars'):
                 config.add_section('user_vars')
@@ -250,11 +236,19 @@ def create_config(args, configfile):
     if not config.has_section('paths'):
         config.add_section('paths')
 
+    if not config.has_section('templates'):
+        config.add_section('templates')
+
     if not configfile:
         default = os.getenv('ZKG_DEFAULT_SOURCE', ZKG_DEFAULT_SOURCE)
         if default:
             config.set('sources', 'zeek', default)
 
+    if not configfile or not config.has_option('templates', 'default'):
+        default = os.getenv('ZKG_DEFAULT_TEMPLATE', ZKG_DEFAULT_TEMPLATE)
+        if default:
+            config.set('templates', 'default', default)
+
     def config_option_set(config, section, option):
         return config.has_option(section, option) and config.get(section,
                                                                  option)
@@ -276,6 +270,8 @@ def create_config(args, configfile):
                             os.path.join(state_dir, 'script_dir'))
     plugin_dir = get_option(config, 'paths', 'plugin_dir',
                             os.path.join(state_dir, 'plugin_dir'))
+    bin_dir = get_option(config, 'paths', 'bin_dir',
+                         os.path.join(state_dir, 'bin'))
     zeek_dist = get_option(config, 'paths', 'zeek_dist', '')
 
     if not zeek_dist:
@@ -284,6 +280,7 @@ def create_config(args, configfile):
     config.set('paths', 'state_dir', state_dir)
     config.set('paths', 'script_dir', script_dir)
     config.set('paths', 'plugin_dir', plugin_dir)
+    config.set('paths', 'bin_dir', bin_dir)
     config.set('paths', 'zeek_dist', zeek_dist)
 
     def expand_config_values(config, section):
@@ -293,6 +290,7 @@ def create_config(args, configfile):
 
     expand_config_values(config, 'sources')
     expand_config_values(config, 'paths')
+    expand_config_values(config, 'templates')
 
     for key, value in config.items('paths'):
         if value and not os.path.isabs(value):
@@ -340,6 +338,7 @@ def create_manager(args, config):
     state_dir = config.get('paths', 'state_dir')
     script_dir = config.get('paths', 'script_dir')
     plugin_dir = config.get('paths', 'plugin_dir')
+    bin_dir = config.get('paths', 'bin_dir')
     zeek_dist = config.get('paths', 'zeek_dist')
 
     if state_dir == default_state_dir():
@@ -348,7 +347,8 @@ def create_manager(args, config):
 
     try:
         manager = zeekpkg.Manager(state_dir=state_dir, script_dir=script_dir,
-                                 plugin_dir=plugin_dir, zeek_dist=zeek_dist)
+                                 plugin_dir=plugin_dir, bin_dir=bin_dir,
+                                 zeek_dist=zeek_dist)
     except (OSError, IOError) as error:
         if error.errno == errno.EACCES:
             print_error('{}: {}'.format(type(error).__name__, error))
@@ -363,7 +363,8 @@ def create_manager(args, config):
 
             permissions_trouble = not all([check_permission(state_dir),
                                            check_permission(script_dir),
-                                           check_permission(plugin_dir)])
+                                           check_permission(plugin_dir),
+                                           check_permission(bin_dir)])
 
             if permissions_trouble and not args.user:
                 print_error(
@@ -599,7 +600,7 @@ def cmd_install(manager, args, config, configfile):
 
     package_infos += new_pkgs
 
-    prompt_for_user_vars(manager, config, configfile, args.force,
+    prompt_for_user_vars(manager, config, configfile, args,
                          [info for info, _, _ in package_infos])
 
     if not args.skiptests:
@@ -923,7 +924,7 @@ def cmd_unbundle(manager, args, config, configfile):
         if not confirmation_prompt('Proceed?'):
             return
 
-    prompt_for_user_vars(manager, config, configfile, args.force,
+    prompt_for_user_vars(manager, config, configfile, args,
                          [info for _, _, info in bundle_info])
 
     error = manager.unbundle(args.bundle_filename)
@@ -1130,7 +1131,7 @@ def cmd_refresh(manager, args, config, configfile):
                           for i in manager.source_packages()}
 
         if src_pkgs_before == src_pkgs_after:
-            print('\tNo changes')
+            print('\tNo membership changes')
         else:
             print('\tChanges:')
             diff = src_pkgs_before.symmetric_difference(src_pkgs_after)
@@ -1297,7 +1298,7 @@ def cmd_upgrade(manager, args, config, configfile):
         if not confirmation_prompt('Proceed?'):
             return
 
-    prompt_for_user_vars(manager, config, configfile, args.force,
+    prompt_for_user_vars(manager, config, configfile, args,
                          [info for info, _, _ in allpkgs])
 
     if not args.skiptests:
@@ -1908,12 +1909,13 @@ def cmd_autoconfig(manager, args, config, configfile):
     dist_option = '--zeek_dist' if have_zeek_config else '--bro_dist'
 
     cmd = subprocess.Popen([zeek_config,
-                            '--site_dir', '--plugin_dir', dist_option],
+                            '--site_dir', '--plugin_dir', '--prefix', dist_option],
                            stdout=subprocess.PIPE, stderr=subprocess.STDOUT,
                            bufsize=1, universal_newlines=True)
 
     script_dir = read_zeek_config_line(cmd.stdout)
     plugin_dir = read_zeek_config_line(cmd.stdout)
+    bin_dir = os.path.join(read_zeek_config_line(cmd.stdout), 'bin')
     zeek_dist = read_zeek_config_line(cmd.stdout)
 
     if configfile:
@@ -1948,6 +1950,8 @@ def cmd_autoconfig(manager, args, config, configfile):
                         script_dir, config_file_exists)
     change_config_value(config, 'paths', 'plugin_dir',
                         plugin_dir, config_file_exists)
+    change_config_value(config, 'paths', 'bin_dir',
+                        bin_dir, config_file_exists)
     change_config_value(config, 'paths', 'zeek_dist',
                         zeek_dist, config_file_exists)
 
@@ -2006,11 +2010,135 @@ def cmd_env(manager, args, config, configfile):
         print(u'setenv BRO_PLUGIN_PATH {}'.format(':'.join(pluginpaths)))
         print(u'setenv ZEEKPATH {}'.format(':'.join(zeekpaths)))
         print(u'setenv ZEEK_PLUGIN_PATH {}'.format(':'.join(pluginpaths)))
+        print(u'setenv PATH {}:$PATH'.format(manager.bin_dir))
     else:
         print(u'export BROPATH={}'.format(':'.join(zeekpaths)))
         print(u'export BRO_PLUGIN_PATH={}'.format(':'.join(pluginpaths)))
         print(u'export ZEEKPATH={}'.format(':'.join(zeekpaths)))
         print(u'export ZEEK_PLUGIN_PATH={}'.format(':'.join(pluginpaths)))
+        print(u'export PATH={}:$PATH'.format(manager.bin_dir))
+
+
+def cmd_create(manager, args, config, configfile):
+
+    tmplname = (args.template
+                or config.get('templates', 'default', fallback=None)
+                or ZKG_DEFAULT_TEMPLATE)
+    try:
+        tmpl = Template.load(config, tmplname, args.version)
+    except LoadError as error:
+        msg = 'problem while loading template {}: {}'.format(tmplname, error)
+        zeekpkg.LOG.exception(msg)
+        print_error('error: ' + msg)
+        sys.exit(1)
+
+    try:
+        package = tmpl.package()
+        uvars = tmpl.define_user_vars()
+        uvar_names = set(package.needed_user_vars())
+
+        # Overlay any requested features onto the package.
+        if args.features:
+            # If the user provided comma-separated values, split the
+            # strings. (Argparse expects space separation.) Also
+            # filter any duplicates.
+            fnames = set()
+            for feat in args.features:
+                fnames |= set([f.strip() for f in feat.split(',') if f])
+            features = tmpl.features()
+            for feature in features:
+                if feature.name() in fnames:
+                    package.add_feature(feature)
+                    uvar_names |= set(feature.needed_user_vars())
+                    fnames.remove(feature.name())
+            if len(fnames) > 0:
+                # Alert if the user requested an unknown feature.
+                knowns = ', '.join([f'"{f.name()}"' for f in features])
+                unknowns = ', '.join([f'"{name}"' for name in fnames])
+                print_error('error: the following features are unknown: {}.'
+                            ' Template "{}" offers {}.'.format(
+                                unknowns, tmpl.name(), knowns or 'no features'))
+                sys.exit(1)
+
+        # Remove user vars we don't actually require from consideration
+        uvars = [uvar for uvar in uvars if uvar.name() in uvar_names]
+
+        # Resolve the variables via user input, args, etc
+        for uvar in uvars:
+            try:
+                uvar.resolve(tmpl.name(), config, args.user_var, args.force)
+            except ValueError:
+                print_error(str.format('error: could not determine value of user variable "{}",'
+                                       ' provide via environment or --user-var', uvar.name()))
+                sys.exit(1)
+
+        # Apply them to the template. After this, any parameter can be
+        # retrieved from the template via tmpl.lookup_param().
+        tmpl._set_user_vars(uvars)
+
+        # Verify that resulting template parameters are formatted correctly
+        try:
+            package.do_validate(tmpl)
+        except zeekpkg.template.InputError as error:
+            print_error('error: template input invalid, ' + str(error))
+            sys.exit(1)
+
+        # And finally, instantiate the package.
+        try:
+            if os.path.isdir(args.packagedir):
+                if not args.force:
+                    print('Package directory {} already exists.'
+                          .format(args.packagedir))
+                    if not confirmation_prompt('Delete?'):
+                        sys.exit(1)
+                try:
+                    delete_path(args.packagedir)
+                    zeekpkg.LOG.info('Removed existing package directory %s', args.packagedir)
+                except OSError as err:
+                    print_error('error: could not remove package directory {}: {}'
+                                .format(args.packagedir, err))
+                    sys.exit(1)
+
+            package.do_instantiate(tmpl, args.packagedir, args.force)
+        except zeekpkg.template.OutputError as error:
+            print_error('error: template instantiation failed, ' + str(error))
+            sys.exit(1)
+    except Exception as error:
+        msg = 'problem during template instantiation: {}'.format(error)
+        zeekpkg.LOG.exception(msg)
+        print_error('error: ' + msg)
+        sys.exit(1)
+
+
+def cmd_template_info(manager, args, config, configfile):
+    tmplname = (args.template
+                or config.get('templates', 'default', fallback=None)
+                or ZKG_DEFAULT_TEMPLATE)
+
+    try:
+        tmpl = Template.load(config, tmplname, args.version)
+    except LoadError as error:
+        msg = 'problem while loading template {}: {}'.format(tmplname, error)
+        zeekpkg.LOG.exception(msg)
+        print_error('error: ' + msg)
+        sys.exit(1)
+
+    tmplinfo = tmpl.info()
+
+    if args.json:
+        print(json.dumps(tmplinfo, indent=args.jsonpretty, sort_keys=True))
+    else:
+        print('API version: ' + tmplinfo['api_version'])
+        print('features: ' + ', '.join(tmplinfo['features']))
+        print('origin: ' + tmplinfo['origin'])
+        print('provides package: ' + str(tmplinfo['provides_package']).lower())
+        print('user vars:')
+        for uvar_name, uvar_info in tmplinfo['user_vars'].items():
+            print('\t{}: {}, {}, used by {}'.format(
+                uvar_name, uvar_info['description'],
+                uvar_info['default'] or 'no default',
+                ', '.join(uvar_info['used_by']) or 'not used'))
+        print('versions: ' + ', '.join(tmplinfo['versions']))
 
 
 class BundleHelpFormatter(argparse.ArgumentDefaultsHelpFormatter):
@@ -2035,8 +2163,10 @@ def top_level_parser():
         '    ``ZKG_CONFIG_FILE``:\t'
         'Same as ``--configfile`` option, but has less precedence.\n'
         '    ``ZKG_DEFAULT_SOURCE``:\t'
-        'The default package source to use (normally {}).'
-        .format(ZKG_DEFAULT_SOURCE)
+        'The default package source to use (normally {}).\n'
+        '    ``ZKG_DEFAULT_TEMPLATE``:\t'
+        'The default package template to use (normally {}).\n'
+        .format(ZKG_DEFAULT_SOURCE, ZKG_DEFAULT_TEMPLATE)
     )
     top_parser.add_argument('--version', action='version',
                             version='%(prog)s ' + zeekpkg.__version__)
@@ -2071,6 +2201,23 @@ def argparser():
                     ' "alice/bro-pkg.index", the following' \
                     ' names work: "foo", "alice/foo", "zeek/alice/foo".'
 
+    def add_uservar_args(parser, force_help=None):
+        parser.add_argument(
+            '--force', action='store_true',
+            help=force_help or "Don't prompt for confirmation or user variables.")
+        parser.add_argument(
+            '--user-var', action='append', metavar='NAME=VAL', type=UserVar.parse_arg,
+            help='A user variable assignment. This avoids prompting'
+            ' for input and lets you provide a value when using --force.'
+            ' Use repeatedly as needed for multiple values.')
+
+    def add_json_args(parser, help_text):
+        parser.add_argument('--json', action='store_true', help=help_text)
+        parser.add_argument(
+            '--jsonpretty', type=int, default=None, metavar='SPACES',
+            help='Optional number of spaces to indent for pretty-printed'
+            ' JSON output.')
+
     top_parser = top_level_parser()
     command_parser = top_parser.add_subparsers(
         title='commands', dest='command',
@@ -2110,9 +2257,6 @@ def argparser():
     sub_parser.set_defaults(run_cmd=cmd_install)
     sub_parser.add_argument(
         'package', nargs='+', help=pkg_name_help)
-    sub_parser.add_argument(
-        '--force', action='store_true',
-        help='Skip the confirmation prompt.')
     sub_parser.add_argument(
         '--skiptests', action='store_true',
         help='Skip running unit tests for packages before installation.')
@@ -2134,6 +2278,7 @@ def argparser():
         ' The default for other cases is to use'
         ' the latest version tag, or if a package has none,'
         ' the default branch, like "main" or "master".')
+    add_uservar_args(sub_parser)
 
     # bundle
     sub_parser = command_parser.add_parser(
@@ -2196,13 +2341,11 @@ def argparser():
         'bundle_filename',
         metavar='filename.bundle',
         help='The path of the bundle file to install.')
-    sub_parser.add_argument(
-        '--force', action='store_true',
-        help='Skip the confirmation prompt.')
     sub_parser.add_argument(
         '--replace', action='store_true',
         help='Using this flag first removes all installed packages before then'
         ' installing the packages from the bundle.')
+    add_uservar_args(sub_parser)
 
     # remove
     sub_parser = command_parser.add_parser(
@@ -2276,9 +2419,6 @@ def argparser():
     sub_parser.set_defaults(run_cmd=cmd_upgrade)
     sub_parser.add_argument(
         'package', nargs='*', default=[], help=pkg_name_help)
-    sub_parser.add_argument(
-        '--force', action='store_true',
-        help='Skip the confirmation prompt.')
     sub_parser.add_argument(
         '--skiptests', action='store_true',
         help='Skip running unit tests for packages before installation.')
@@ -2290,6 +2430,7 @@ def argparser():
     sub_parser.add_argument(
         '--nosuggestions', action='store_true',
         help='Skip automatically installing suggested packages.')
+    add_uservar_args(sub_parser)
 
     # load
     sub_parser = command_parser.add_parser(
@@ -2410,13 +2551,7 @@ def argparser():
         '--nolocal', action='store_true',
         help='Do not read information from locally installed packages.'
         ' Instead read info from remote GitHub.')
-    sub_parser.add_argument(
-        '--json', action='store_true',
-        help='Output package information as JSON.')
-    sub_parser.add_argument(
-        '--jsonpretty', type=int, default=None, metavar='SPACES',
-        help='Optional number of spaces to indent for pretty-printed'
-        ' JSON output.')
+    add_json_args(sub_parser, 'Output package information as JSON.')
     sub_parser.add_argument(
         '--allvers', action='store_true',
         help='When outputting package information as JSON, show metadata for'
@@ -2439,7 +2574,7 @@ def argparser():
     sub_parser.add_argument(
         'config_param', nargs='?', default='all',
         choices=['all', 'sources', 'user_vars', 'state_dir', 'script_dir',
-                 'plugin_dir', 'zeek_dist', 'bro_dist'],
+                 'plugin_dir', 'bin_dir', 'zeek_dist', 'bro_dist'],
         help='Name of a specific config file field to output.')
 
     # autoconfig
@@ -2476,6 +2611,64 @@ def argparser():
         formatter_class=argparse.ArgumentDefaultsHelpFormatter)
     sub_parser.set_defaults(run_cmd=cmd_env)
 
+    # create
+    sub_parser = command_parser.add_parser(
+        'create', help='Create a new Zeek package.',
+        description='This command creates a new Zeek package in the directory'
+        ' provided via --packagedir. If this directory exists, zkg will not'
+        ' modify it unless you provide --force.',
+        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
+    sub_parser.add_argument(
+        '--packagedir', metavar='DIR', required=True,
+        help='Output directory into which to produce the new package. Required.')
+    sub_parser.add_argument(
+        '--version',
+        help='The template version to use.  A version tag, branch name, or'
+        ' commit hash may be specified here.  If --template refers to a local'
+        ' git repo with a working tree, then zkg uses it as-is and the version'
+        ' is ignored.  The default for other cases is to use the latest'
+        ' version tag, or if a template has none, the default branch, like'
+        ' "main" or "master".')
+    sub_parser.add_argument(
+        '--features', nargs='+', metavar='FEATURE',
+        help='Additional features to include in your package. Use the ``template'
+        ' info`` command for information about available features.')
+    sub_parser.add_argument(
+        '--template', metavar='URL',
+        help='By default, zkg uses its own package template. This makes it'
+        ' select an alternative.')
+    sub_parser.set_defaults(run_cmd=cmd_create)
+    add_uservar_args(sub_parser)
+
+    # Template management commands
+
+    sub_parser = command_parser.add_parser(
+        'template', help='Manage package templates.')
+
+    template_command_parser = sub_parser.add_subparsers(
+        title='template commands',
+        help='See `%(prog)s template <command> -h for per-command usage info.')
+
+    # template info
+
+    sub_parser = template_command_parser.add_parser(
+        'info', help='Shows information about a package template.',
+        description='This command shows versions and supported features for'
+        ' a given package.')
+    add_json_args(sub_parser, 'Output template information as JSON.')
+    sub_parser.add_argument(
+        '--version',
+        help='The template version to report on.  A version tag, branch name,'
+        ' or commit hash may be specified here.  If the selected template'
+        ' refers to a local git repo, the version is ignored.  The default'
+        ' for other cases is to use the latest version tag, or if a template'
+        ' has none, the default branch, like "main" or "master".')
+    sub_parser.add_argument(
+        'template', metavar='URL', nargs='?',
+        help='URL of a package template repository, or local path to one.'
+        ' When not provided, the configured default template is used.')
+    sub_parser.set_defaults(run_cmd=cmd_template_info)
+
     return top_parser
 
 
